import numpy as np

from re import match
from typing import List

from .classes import Reaction, Nuclide, Nuclib
from .xslib import XSlib
from ..constants import MT_dict
from ..helpers import parse_filelines


class MGXSlib(Nuclib):

    def __init__(self, nuclides: List[Nuclide] = []):
        super().__init__("mgxs", nuclides)

    def to_basic_xslib(self, fluxes: List[float], original_xslib: XSlib = None):
        xslib = XSlib()
        for nuclide in self:
            xslib.append(Nuclide(nuclide.name))
            for reaction in nuclide:
                xses = (reaction.mgxs * fluxes[reaction.start_group-1:]).sum() / fluxes.sum()
                Qvalue = original_xslib(nuclide.name)(reaction.name).Qvalue if original_xslib else 0
                xslib[-1].append(Reaction(reaction.name, data={"xses": xses, "Qvalue": Qvalue}))
        return xslib

    def to_fixed_xslib(self, flux: List[float]):
        xslib = XSlib(with_burnup=True)
        xslib.burnups = [0, 100]
        for nuclide in self:
            xslib.append(Nuclide(nuclide.name))
            for reaction in nuclide:
                xses = (reaction.mgxs * flux[reaction.start_group-1:]).sum() / flux.sum()
                xslib[-1].append(Reaction(reaction.name, data={"xses": [xses, xses]}))
        return xslib

    def to_unfixed_xslib(self, fluxes: List[List[float]], burnups: List[float]):
        xslib = XSlib(with_burnup=True)
        xslib.burnups = burnups
        for nuclide in self:
            xslib.append(Nuclide(nuclide.name))
            for reaction in nuclide:
                xses = (reaction.mgxs[np.newaxis, :] * fluxes[:, reaction.start_group-1:]).sum(axis=1) / fluxes.sum(axis=1)
                xslib[-1].append(Reaction(reaction.name, data={"xses": xses}))
        return xslib

    @classmethod
    def from_datfile(cls, filepath: str):

        mgxslib = cls()
        with open(filepath, 'r') as fileopen:
            filelines = fileopen.readlines()

        index = parse_filelines(filelines, 'NucId') + 1
        for fileline in filelines[index:]:
            if results := match(r"^(\d+)\s+([A-z0-9]+)\s+(\d+)$", fileline):
                name = results.group(2)
                mgxslib.append(Nuclide(name))
            elif results := match(r"^\s+(\d+)\s+(\d+)\s+([0-9E\-\.]+)", fileline):
                MT = int(results.group(1))
                start_group = int(results.group(2))
                xses = [float(xs) for xs in fileline.split()[2:]]
                mgxslib[-1].append(Reaction(MT_dict[MT], data={"start_group": start_group, "mgxs": np.array(xses)}))
        return mgxslib



import numpy as np
import matplotlib.pyplot as plt

from typing import Literal, List
from copy import deepcopy
from time import mktime, strptime, struct_time

from . import model


class Histogram:

    def __init__(self, times: List[str], powers: List[float], mode: Literal['constpower', 'constflux'] = 'constpower', model: model.Model = None):
        self.times = times
        self.powers = np.array(powers)
        self.model = model
        self.mode = mode

    def __add__(self, histogram):
        return Histogram(self.times + histogram.times, np.concatenate([self.powers, histogram.powers]))

    def __mul__(self, ratio):
        return Histogram(self.times, self.powers * ratio)

    def __truediv__(self, ratio):
        return Histogram(self.times, self.powers / ratio)

    @property
    def stamps(self):
        return np.array([mktime(strptime(time, "%Y-%m-%d %H:%M")) for time in self.times])

    @property
    def time_lengths(self):
        return self.stamps[1:] - self.stamps[:-1]

    @property
    def average_powers(self):
        return (self.powers[:-1, 1] + self.powers[1:, 0]) / 2

    @property
    def burnup(self):
        return sum(self.time_lengths * self.average_powers)

    @burnup.setter
    def burnup(self, burnup):
        self.powers *= burnup / self.burnup

    def copy(self):
        return deepcopy(self)

    def get_power(self, time) -> float:
        if isinstance(time, struct_time):
            time = mktime(time)
        if (time <= self.stamps[0]) or (time >= self.stamps[-1]):
            return 0
        else:
            index = np.argwhere(self.stamps < time)[-1][0]
            init_time = self.stamps[index]
            final_time = self.stamps[index+1]
            init_power = self.powers[index][1]
            final_power = self.powers[index+1][0]
            power = (time - init_time) / (final_time - init_time) * (final_power - init_power) + init_power  # linear interpolation
            return power

    def plot_power(self, time_length=10, axes: plt.Axes = None) -> plt.Axes:
        if axes is None:
            _, axes = plt.subplots(figsize=(20, 3))
        times = np.linspace(self.stamps[0], self.stamps[-1], int((self.stamps[-1] - self.stamps[0])/(86400*time_length)))
        powers = [self.get_power(time) for time in times]
        axes.plot(times, powers, ".-", markersize=10)
        axes.set_xticks(self.stamps, self.times, rotation=90)
        axes.set_ylim(0, )
        return axes

    def to_model(self, step_length: float = 10 * 86400):

        self.model.remove_node("burnup")
        for i in range(len(self.stamps)-1):
            repeat = int((self.stamps[i+1]-self.stamps[i]) / step_length) + 1
            step_length_exact = (self.stamps[i+1]-self.stamps[i]) / repeat
            if self.powers[i][1] != 0 or self.powers[i+1][0] != 0:
                if self.powers[i][1] == self.powers[i+1][0]:
                    self.model.add_burnup(mode=self.mode,
                                          time=step_length_exact,
                                          val=self.powers[i][1],
                                          repeat=repeat,
                                          unit='second')
                else:
                    for j in range(repeat):
                        self.model.add_burnup(mode=self.mode,
                                              time=step_length_exact,
                                              val=self.get_power(self.stamps[i]+step_length_exact*(j+1/2)),
                                              unit='second')
            else:
                self.model.add_burnup(mode='decay',
                                      time=step_length_exact,
                                      val=0,
                                      repeat=repeat,
                                      unit='second')
        return self.model



'''
Author: albertzhang albert.zhangweij@outlook.com
Date: 2023-12-22 11:20:00
Description: 

Copyright (c) 2023 by THU-RSAG, All Rights Reserved. 
'''
from .isolib import *
from .xslib import *
from .mgxslib import *
from .fylib import *
from .decaylib import *


'''
Author: albertzhang albert.zhangweij@outlook.com
Date: 2023-12-20 11:56:37
Description: 

Copyright (c) 2023 by THU-RSAG, All Rights Reserved. 
'''
from re import match
from typing import List

from .classes import Nuclide, Nuclib, DecayReaction
from .xslib import XSlib
from ..helpers import parse_filelines


class DECAYlib(Nuclib):
    
    def __init__(self, nuclides: List[Nuclide] = []):
        super().__init__("decay", nuclides)

    @classmethod
    def from_datfile(cls, filepath: str):

        decaylib = cls()
        with open(filepath, 'r') as fileopen:
            filelines = fileopen.readlines()

        index = parse_filelines(filelines, 'NUIT decay library') + 1
        for fileline in filelines[index:]:
            if results := match(r"^(\d+)\s+([A-z0-9]+)\s+(\d+)\s+([0-9E\+\-\.]+)\s+([0-9E\+\-\.]+)\s+([0-9E\+\-\.]+)\s+([0-9E\+\-\.]+)", fileline):
                name = results.group(2)
                decaylib.append(Nuclide(name, data={"rate":float(results.group(4)), "Qalpha":float(results.group(5)),
                                                     "Qbeta":float(results.group(6)), "Qgamma":float(results.group(7))}))
            elif results := match(r"^\s+(\d+)\s+([0-9E\+\-\.]+)", fileline):
                decayMT = int(results.group(1))
                ratio = float(results.group(2))
                # decaylib[-1].append({'decayMT': decayMT, 'ratio': ratio})
                decaylib[-1].append(DecayReaction(decayMT, data={'ratio': ratio}))
        return decaylib


'''
Author: albertzhang albert.zhangweij@outlook.com
Date: 2023-12-22 11:20:00
Description: 

Copyright (c) 2023 by THU-RSAG, All Rights Reserved. 
'''
import numpy as np

from re import match
from typing import List

from .classes import Reaction, Nuclide, Nuclib
from ..constants import MT_dict
from ..helpers import parse_filelines


class XSlib(Nuclib):

    def __init__(self, nuclides: List[Nuclide] = [], with_burnup: bool = False):
        self.with_burnups = with_burnup
        super().__init__('xses', nuclides)

    @classmethod
    def from_datfile(cls, filepath: str):

        xslib = cls()
        with open(filepath, 'r') as fileopen:
            filelines = fileopen.readlines()

        # read burnup table
        if (index := parse_filelines(filelines, 'BU(MWd/kgHM)')):
            xslib.with_burnups = True
            xslib.burnups = [float(bu) for bu in filelines[index+1].split()]
        else:
            xslib.with_burnups = False
            index = 0

        # count index until the xs table
        index = parse_filelines(filelines, 'NucId', index) + 1
        index_final = parse_filelines(filelines, 'NUIT fission product yield data', index) - 1 if not xslib.with_burnups else len(filelines)
        for fileline in filelines[index: index_final]:
            if (results := match(r'^(\d+)\s+([A-Za-z0-9_]+)\s+(\d+)\s*\n$', fileline)):  # nuclide line
                xslib.append(Nuclide(name=results.group(2)))
            elif results := match(r'^\s+(\d+)\s+([0-9E\-\.\+]+)\s+([0-9E\-\.\+]+)', fileline):  # reaction line
                fullMT = int(results.group(1))  # fullMT = MT + status
                Qvalue = float(results.group(2))
                if xslib.with_burnups:
                    xses = [float(xs) for xs in fileline.split()[1:]]
                else:
                    xses = float(results.group(3))
                xslib[-1].append(Reaction(MT_dict[fullMT], data={"Qvalue": Qvalue, "xses": xses}))
        xslib.nucids = [nuclide.nucid for nuclide in xslib]
        return xslib

    def to_datfile(self, filepath: str):

        fileopen = open(filepath, 'w')
        fileopen.write('*************************** NUIT one-group neutron cross-section data ***************************\n')
        fileopen.write(f'Number of isotopes with neutron data: \n\t{len(self)}\n')
        if self.with_burnups:
            fileopen.write(f'Number of burnup steps:\n\t{len(self.burnups)}\n\n')
        else:
            fileopen.write('Number of energy groups:\n\t361\n\n')

        if self.with_burnups:
            fileopen.write('BU(MWd/kgHM)\n')
            fileopen.write('   '.join([f'{burnup:<12.8E}' for burnup in self.burnups]) + '\n\n')

        if self.with_burnups:
            fileopen.write('NucId    NucName  MT\n')
            for nuclide in self:
                fileopen.write(f'{nuclide.nucid:<8d} {nuclide.name:<8s} {nuclide.num_rec:<8d}' + '\n')
                for reaction in nuclide:
                    fileopen.write(f'                  {reaction.MT:<6d}' + '   '.join([f"{xs:<12.6E}" for xs in reaction.xses]) + '\n')
        else:
            fileopen.write('NucId    NucName  MT     Qvalue         Cross section\n')
            for nuclide in self:
                fileopen.write(f'{nuclide.nucid:<8d} {nuclide.name:<8s} {nuclide.num_rec:<8d}' + '\n')
                for reaction in nuclide:
                    fileopen.write(f'                  {reaction.MT:<6d}{reaction.Qvalue:>13.6E}   {reaction.xses:<12.6E}' + '\n')




'''
Author: albertzhang albert.zhangweij@outlook.com
Date: 2023-12-22 11:20:00
Description: 

Copyright (c) 2024 by THU-RSAG, All Rights Reserved. 
'''
from .constants import *
from .helpers import *
from .model import *
from .output import *
from .histogram import *

from . import data
from . import beta



from typing import List, Literal
from re import match
from .constants import time_convert_units, Atomic_dict, Atomic_list

def parse_filelines(filelines: List[str], target: str, start_index: int = 0):
    index = start_index
    filelen = len(filelines)
    while target not in filelines[index]:
        index += 1
        if index == filelen:
            return None
    return index

def convert_time(time, unit1, unit2):
    return time * time_convert_units[unit1] / time_convert_units[unit2]


def decompose_nucname(nucname, style: Literal['NUIT', 'OpenMC'] = 'NUIT'):
    if style == 'NUIT':
        pattern = r'^([a-zA-Z]+)(\d+)_?(m[123])?$'
    elif style == 'OpenMC':
        pattern = r'^([a-zA-Z]+)(\d+)?(m)?$'
    match_results = match(pattern, nucname)

    if match_results is None:
        return None
    else:
        match_results = match_results.groups()

    symbol = match_results[0]
    mass = int(match_results[1])
    state = 0 if match_results[2] is None else int(match_results[2][1])
    atomic = Atomic_dict[symbol]
    nucid = atomic * 10000 + mass * 10 + state

    nuc_info = {'symbol': symbol, 'mass': mass, 'state': state, 'atomic': atomic, 'nucid': nucid}
    return nuc_info

def convert_nucid_to_name(nucid: int):
    atomic = nucid // 10000
    mass = nucid // 10 % 1000
    state = nucid % 10
    symbol = Atomic_list[atomic]
    name = f'{symbol}{mass}'
    if state != 0:
        name += f'm{state}'
    return name



import platform
import time
import subprocess
import xml.etree.ElementTree as et
import xml.dom.minidom as mdm

from re import match as re_match
from sys import stdout
from typing import Literal, List, Union

from . import constants
from . import output


class Model(et.ElementTree):
    
    def __init__(self, filepath: str = None):
        super().__init__(element=et.Element('input'))
        self._root = self.getroot()
        if filepath is not None:
            self.parse(filepath)

    def get_node(self, node_name: str) -> et.Element:
        """
        Get the node with the specified name from the xml tree.

        Args:
            node_name (str): The name of the node to be retrieved.

        Returns:
            et.Element: The node with the specified name.
        """
        return self._root.find(node_name)

    def remove_node(self, node_name: str) -> None:
        """
        Remove any node with the specified name from the xml tree.

        Args:
            node_name (str): The name of the node to be removed.
        """
        if (node := self._root.find(node_name)) is not None:
            self._root.remove(node)

    def set_library(self,
                    lib_path: str,
                    only_use_endf_data: bool = False,
                    sub_xs_lib: str = None,
                    xs_lib: str = None,
                    ) -> None:
        """
        Set the library information.

        Args:
            lib_path (str): The path to the library.
            sub_xs_lib (str, optional): The path to the sub XS library. Defaults to None.
            only_use_endf_data (bool, optional): Flag indicating whether to only use ENDF data. Defaults to False.
            xs_lib (str, optional): The path to the XS library. Defaults to None.
        """
        # remove the existing library node
        if (node := self._root.find('library')) is not None:
            self._root.remove(node)

        # collect all the input information
        info = {'lib_path': str(lib_path),
                'only_use_endf_data': str(int(only_use_endf_data))}
        if sub_xs_lib:
            info['sub_xs_lib'] = str(sub_xs_lib)
        if xs_lib:
            info['xs_lib'] = str(xs_lib)
        et.SubElement(self._root, 'library', info)

    def add_nuclide(self,
                    name: str,
                    den: float,
                    unit: Literal['gram', 'barn'] = 'gram') -> None:
        """
        Add a nuclide to the material.

        Args:
            name (str): The name of the nuclide.
            den (float): The density of the nuclide.
            unit (Literal['gram', 'barn'], optional): The unit of the density. Defaults to 'gram'.
        """

        # add the material node if it does not exist
        if self._root.find('material') is None:
            et.SubElement(self._root, 'material', {'unit': unit})
        material = self._root.find('material')

        # input validity check
        if material.attrib['unit'] != unit:
            raise ValueError('Unit of material is not consistent.')

        # add the nuclide node
        mres = re_match(r'([A-z]*)(\d*)([m]?)', name)
        name, mass, status = mres.groups()
        status = 1 if status == 'm' else 0
        A = constants.Atomic_list.index(name)+1
        id = int(A)*10000+int(mass)*10 + status

        # collect all the input information
        info = {'den': str(den), 'id': str(id)}
        et.SubElement(material, 'nuc', info)

    def add_burnup(self,
                    time: float,
                    val: float,
                    repeat: int = 1,
                    mode: Literal['constpower', 'constflux', 'decay'] = 'constpower',
                    unit: Literal['day', 'hour', 'minute', 'second'] = 'day',
                    spectrum: List[float] = None):
        """
        Adds a burnup node with the specified time, value, and other optional parameters to the XML tree.

        Args:
            time (float): The time value for the burnup node.
            val (float): The value for the burnup node (power or flux).
            repeat (int, optional): The number of times to repeat the burnup node (default: 1).
            mode (Literal['constpower', 'constflux', 'decay'], optional): The mode of the burnup node (default: 'constpower').
            unit (Literal['day', 'hour', 'minute', 'second'], optional): The unit of the time value (default: 'day').
            spectrum (List[float], optional): The spectrum values for the burnup node (default: None).
        """
        # add the burnup node if it does not exist
        if self._root.find('burnup') is None:
            et.SubElement(self._root, 'burnup', {})
        burnup = self._root.find('burnup')

        # input validity check
        if unit not in ['day', 'hour', 'minute', 'second']:
            raise ValueError('Time unit should be day, hour, minute or second.')
        if mode not in ['constpower', 'constflux', 'decay']:
            raise ValueError('Power mode should be constpower, constflux or decay.')

        # add the burn node
        # collect all the input information
        val_name = 'power' if mode == 'constpower' else 'flux'
        info = {'time': str(time), 'unit': unit,
                'mode': mode, val_name: str(val),
                'repeat': str(repeat)}
        burnup_node = et.SubElement(burnup, 'burn', info)

        # add the spectrum if necessary
        if spectrum is not None:
            spectrum = ' '.join([f"{s:8.3e}" for s in spectrum])
            burnup_node.set('spectrum', spectrum)

    def set_gen_library(self,
                        flux_spectrum,
                        energy_boundary,
                        temperature,
                        data_path,
                        decay_file='decay',
                        neutron_file='neutron',
                        nfy_file='nfy',
                        output_path='NUITLIB'):

        raise KeyError("Method deprecated!")

        # modify the flux_spectrum into str
        flux_spectrum = ' '.join([str(s) for s in flux_spectrum])
        if energy_boundary[0] < 1E-6:
            energy_boundary += 1E-6
        energy_boundary = ' '.join([str(s) for s in energy_boundary])

        # remove the existing library node
        if (node := self._root.find('gen_library')) is not None:
            self._root.remove(node)

        info = {}
        for item in ['flux_spectrum', 'temperature', 'energy_boundary',
                     'data_path', 'decay_file', 'neutron_file', 'nfy_file', 'output_path']:
            if locals()[item] is not None:
                info[item] = str(locals()[item])
        et.SubElement(self._root, 'gen_library', info)

    def set_output(self,
                   type: Literal['isotope', 'gammaspectra', 'absorption', 'fission', 'decayheat',
                                 'decayheat-lp', 'decayheat-em', 'decayheat-hp', 'radioactivity', 'betaspectra'],
                   print_all_step: bool = True,
                   integral: bool = False) -> None:
        """
        Set the type of output, whether to print all steps, and whether to calculate the integral.

        Args:
            type (Literal['isotope', 'gammaspectra', 'absorption', 'fission', 'decayheat',
                        'decayheat-lp', 'decayheat-em', 'decayheat-hp', 'radioactivity', 'betaspectra']): The type of output.
            print_all_step (bool): Whether to print all steps. Default is True.
            integral (bool): Whether to calculate the integral. Default is False.
        """
        # add the output node if it does not exist
        if self._root.find('output') is None:
            et.SubElement(self._root, 'output', {})
        output = self._root.find('output')

        # input validity check
        if type not in ['isotope', 'gammaspectra', 'absorption', 'fission', 'decayheat',
                        'decayheat-lp', 'decayheat-em', 'decayheat-hp', 'radioactivity', 'betaspectra']:
            raise KeyError('Not a valid output type.')

        # add the output node
        # collect all the input information
        info = {'type': type,
                'print_all_step': str(int(print_all_step)),
                'integral': str(int(integral))}
        et.SubElement(output, 'table', info)

    def export_to_xml(self, filepath: str) -> None:
        """
        Export the object to an XML file.

        Args:
            filepath (str): The path to the XML file.
        """
        super().write(file_or_filename=filepath)

        # additional process to make the xml file pretty
        dom = mdm.parse(filepath)
        pdom = dom.toprettyxml(indent='\t', newl='\n')
        dom.unlink()
        with open(filepath, 'w') as fileopen:
            fileopen.write(pdom)

    def __call__(self, filepath: str, nuitpath: Union[str, None] = None, show_on_console: bool=True) -> output.Output:
        """
        Run NUIT with the specified input file.

        Args:
            filepath (str): The path to the input file.
            nuitpath (str): The path to the NUIT executable.

        Returns:
            output.Output: The output of NUIT.
        """
        self.export_to_xml(filepath)
        print(f"""Running NUIT...
|-Executable:{nuitpath}
|-Input:{filepath}
|-Output:{filepath+'.out'}
|-Time:{time.strftime('%Y-%m-%d %H:%M:%S', time.localtime())}""")

        if nuitpath is None:
            if constants.config['nuit_path'] is None:
                raise ValueError('NUIT path not specified.')
            nuitpath = constants.config['nuit_path']

        if platform.system() == 'Windows':
            popen = subprocess.Popen(f"{nuitpath} -i {filepath}",
                                        shell=True,
                                        stdout=subprocess.PIPE,
                                        stderr=subprocess.PIPE)
        else:
            popen = subprocess.Popen(f"wine {nuitpath} -i {filepath}",
                                    shell=True,
                                    stdout=subprocess.PIPE,
                                    stderr=subprocess.PIPE)
        if show_on_console:
            with open(filepath+'.log', 'w') as fileopen:
                while popen.poll() is None:
                    out = popen.stdout.readline().decode()
                    stdout.write(out)
                    fileopen.write(out[:-1])
                else:
                    out = popen.stdout.readline().decode()
                    stdout.write(out)
                    fileopen.write(out[:-1])
                if popen.poll() != 0:
                    err = popen.stderr.readline().decode()
                    stdout.write(err)
                    fileopen.write(err)
        else:
            with open(filepath+'.log', 'w') as fileopen:
                fileopen.write(out.stdout.read().decode())
                fileopen.write(out.stderr.read().decode())
        return output.Output(filepath+'.out')



'''
Author: albertzhang albert.zhangweij@outlook.com
Date: 2023-12-22 11:20:00
Description: 

Copyright (c) 2024 by THU-RSAG, All Rights Reserved. 
'''
from pickle import load
from collections import namedtuple
from typing import Union

from ..helpers import decompose_nucname, convert_nucid_to_name
from ..constants import MT_dict

ISOclause = namedtuple("ISOclause", ["nucid", "MT", "fracm"])
def __repr__(self):
    return f"{convert_nucid_to_name(self.nucid)} {self.MT}: {self.fracm}"
ISOclause.__repr__ = __repr__

class ISOlib(list):

    def __init__(self, clauses: list[ISOclause] = []):
        super().__init__(clauses)

    def __call__(self, nuclide: Union[str, int], reaction: Union[str, int]):
        if isinstance(reaction, str):
            reaction = MT_dict[reaction]
        if isinstance(nuclide, str):
            nuclide = decompose_nucname(nuclide)['nucid']
        clause = next((clause for clause in self if getattr(clause, 'nucid') == nuclide and getattr(clause, 'MT') == reaction))
        return clause

    @classmethod
    def from_pickle(cls, filepath: str):
        
        with open(filepath, 'rb') as fileopen:
            isomerics = load(fileopen)
        isolib = ISOlib([ISOclause(iso['nucid'], iso["MT"], iso["fracm"]) for iso in isomerics])
        return isolib



import numpy as np

from typing import Any, List, Union

from .helpers import parse_filelines


class Output():
    """
    container for NUIT Output file.

    Attributes:
        filelines (list): List of strings representing the lines of the output file.
        nuclide_masses (list): List of dictionaries containing nuclide information, including nuclide ID, name, and data.
        burnups (ndarray): Array of burnup values.
        powers (ndarray): Array of power values.
        timesteps (ndarray): Array of timestep values.
        fluxes (ndarray): Array of flux values.

    Methods:
        __init__(filepath): Initializes the Output object by reading the file and parsing the data.
        _read_isotopes(): Parses the nuclide density data from the file.
        _read_burnups(): Parses the burnup data from the file.
        _read_powers(): Parses the power data from the file.
        _read_timesteps(): Parses the timestep data from the file.
        _read_fluxes(): Parses the flux data from the file.
        match(key, value, type): Matches an element from the stored tables using a key-value pair.
        get_nuclide_mass(nuclide, burnup): Gets the nuclide mass at a certain burnup depth.
        output_nuclide_mass(nuclide_list, burnup): Gets the nuclide mass in batch.
    """

    def __init__(self, filepath) -> None:
        """
        Initialize the Output object.

        Args:
            filepath (str): The path to the file to be read.
        """
        with open(filepath) as fileopen:
            self.filelines = fileopen.readlines()
        
        # to avoid format diffenrence when print_all_steps=0/1
        try:
            self._read_burnups()
            self._read_powers()
            self._read_timesteps()
            self._read_fluxes()
        except:
            self._read_info_two_steps()
            pass
        self._read_isotopes()

    def _read_info_two_steps(self) -> None:
        index = parse_filelines(self.filelines, 'Nuclide Density(n*1.0E+24) Table, Instantaneous', 0) + 3
        self.timesteps = [float(t) for t in self.filelines[index].split()[2::3]]
        self.fluxes = [float(t) for t in self.filelines[index+1].split()[2::3]]
        self.powers = [float(t) for t in self.filelines[index+2].split()[2::3]]
        self.burnups = [float(t) for t in self.filelines[index+3].split()[2::3]]

    def _read_isotopes(self) -> None:
        """
        Extracts the nuclide information, including nuclide ID, name, and data, from the filelines and stores them in the `nuclide_masses` attribute.
        
        Returns:
            None
        """
        self.nuclide_masses = []
        index = parse_filelines(self.filelines, 'Nuclide Density', 0) + 1
        index = parse_filelines(self.filelines, 'NucID', index) + 2
        while "Non-Actinide" not in self.filelines[index]:
            line = self.filelines[index].split()
            nuc_id, nuc_name = int(line[0]), line[1]
            nuc_den = np.asarray([float(item) for item in line[2:]])
            self.nuclide_masses.append({'nucid': nuc_id,
                                        'name': nuc_name,
                                        'data': nuc_den})
            index += 1

    def _read_burnups(self) -> None:
        """
        Extracts the burnup values from the filelines and stores them in the `burnups` attribute.

        Returns:
            None
        """
        self.burnups = []
        ind = parse_filelines(self.filelines, 'BU(MWd/kgHM):', 0)
        line = self.filelines[ind].split()
        for bu in line[1:]:
            self.burnups.append(float(bu))
        self.burnups = np.asarray(self.burnups)

    def _read_powers(self) -> None:
        """
        Reads the power values from the filelines and stores them in the powers attribute.

        Returns:
            None
        """
        self.powers = []
        ind = parse_filelines(self.filelines, 'Power(MW):', 0)
        line = self.filelines[ind].split()
        for pw in line[1:]:
            self.powers.append(float(pw))
        self.powers = np.asarray(self.powers)

    def _read_timesteps(self) -> None:
        """
        Reads the timesteps from the filelines and stores them in the `timesteps` attribute.

        Returns:
            None
        """
        self.timesteps = []
        ind = parse_filelines(self.filelines, 'TotalTime(s):', 0)
        line = self.filelines[ind].split()
        for ts in line[1:]:
            self.timesteps.append(float(ts))
        self.timesteps = np.asarray(self.timesteps)

    def _read_fluxes(self) -> None:
        """
        Reads the fluxes from the filelines and stores them in the 'fluxes' attribute.
        
        Returns:
            None
        """
        self.fluxes = []
        ind = parse_filelines(self.filelines, 'Flux(n/cm^2/s):', 0)
        line = self.filelines[ind].split()
        for fl in line[1:]:
            self.fluxes.append(float(fl))
        self.fluxes = np.asarray(self.fluxes)

    def match(self, key='name', value='Cs137', type='nuclide') -> Any:
        """
        Match an element from stored tables using key-value pair.
        
        Returns:
            Any: Depends on the type of the table.
        """
        if type == 'nuclide':
            table = self.nuclide_masses
        elif type == 'burnup':
            table = self.burnups
        val = next((item for item in table if item[key] == value), None)
        return val

    def get_nuclide_mass(self, nuclide: str, burnup: float = None) -> Union[float, List[float]]:
        """"
        Get nuclide mass at a certain burnup depth.
        
        Returns:
            mass (Union(float, List[float])): Nuclide mass at a certain burnup depth.
        """
        if burnup:
            return self.match('name', nuclide, 'nuclide')[abs(self.burnups-burnup).argmin()]
        else:
            return self.match('name', nuclide)['data']

    def output_nuclide_mass(self, nuclide_list, burnup=None) -> List[float]:
        """
        Get nuclide mass in batch.
        
        Returns:
            mass (List[float]): Nuclide mass in batch.
        """
        return [self.get_nuclide_mass(nuclide, burnup) for nuclide in nuclide_list]



'''
Author: albertzhang albert.zhangweij@outlook.com
Date: 2023-12-22 11:20:00
Description: 

Copyright (c) 2024 by THU-RSAG, All Rights Reserved. 
'''
import setuptools

with open("README.md", "r") as fileopen:
    long_description = fileopen.read()

setuptools.setup(
    name="pyNUIT-NSAG",
    version="0.3",
    author="THU-RSAG",
    author_email="z-wj21@mails.tsinghua.edu.cn",
    description="Wrapper for source term alaysis code NUIT",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/thu-inet/pyNUIT",
    packages=setuptools.find_packages(),
    install_requires=['numpy>=1.26.0', 'pandas>=2.1.3', 'matplotlib>=3.8.2'],
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        ],
)



'''
Author: albertzhang albert.zhangweij@outlook.com
Date: 2023-12-27 16:20:50
Description: 

Copyright (c) 2023 by THU-RSAG, All Rights Reserved. 
'''
from re import match
from collections import namedtuple
from typing import Union
from functools import cached_property

from .classes import Nuclib
from ..helpers import parse_filelines, decompose_nucname, convert_nucid_to_name

FYclause = namedtuple("FYclause", ["nuclide", "target", "group", "energy", "fyield"])
def __repr__(self):
    return f"{convert_nucid_to_name(self.nuclide)}->{convert_nucid_to_name(self.target)} {self.fyield}"
FYclause.__repr__ = __repr__


class FYlib(list):
    
    def __init__(self, clauses: list[FYclause] = []):
        super().__init__(clauses)

    def __call__(self, target: Union[int, str], nuclide: Union[int, str]="U235"):
        if isinstance(target, str):
            target = decompose_nucname(target)['nucid']
        if isinstance(nuclide, str):
            nuclide = decompose_nucname(nuclide)['nucid']
            
        clause = next((clause for clause in self if clause.nuclide == nuclide and clause.target == target), None)
        return clause

    @classmethod
    def from_datfile(cls, filepath: str):
        fylib = cls([])
        with open(filepath, 'r') as fileopen:
            filelines = fileopen.readlines()
        index = parse_filelines(filelines, "NUIT fission product yield data") + 4
        for i in range(index, len(filelines)):
            if (results := match(r"^(\d+)\s+([A-z0-9]+)\s+(\d+)\s+(\d)\s+$", filelines[i])):
                nuclide = decompose_nucname(results.group(2))['nucid']
                group = int(results.group(4))
                energy = [float(val) for val in filelines[i+1].split()]
            elif (results := match(r"^\s+(\d+)\s+([A-z0-9\.\+\-]+)\s+", filelines[i])):
                target = int(results.group(1))
                fyield = [float(val) for val in filelines[i].split()[1:]]
                fylib.append(FYclause(nuclide, target, group, energy, fyield))
        return fylib



'''
Author: albertzhang albert.zhangweij@outlook.com
Date: 2023-12-22 11:20:00
Description: 

Copyright (c) 2023 by THU-RSAG, All Rights Reserved. 
'''

from typing import List, Union, Any
from functools import cached_property
from ..constants import MT_dict, Atomic_list, MT_to_nucid_dict, particle_dict
from ..helpers import decompose_nucname


class Reaction():

    def __init__(self, name: str, data: dict):
        self.name = name
        self.data = data

    @cached_property
    def fullMT(self):
        return MT_dict[self.name]
    
    @cached_property
    def MT(self):
        if 'm1' in self.name:
            return MT_dict[self.name[:-2]]
        return MT_dict[self.name]

    @cached_property
    def status(self):
        return int(self.name[:-2] == 'm1')

    @cached_property
    def varid(self):
        return (MT_to_nucid_dict[self.MT], self.status)

    def __repr__(self):
        return self.name

    def __getattr__(self, attr: str):
        if attr in self.data:
            return self.data[attr]
        else:
            return None


class DecayReaction:
    
    def __init__(self, decayMT: int, data: dict):
        self.decayMT = decayMT
        self.data = data

    @staticmethod
    def help():
        return "DecayReaction(decayMT, data={'ratio':float})"
        
    @cached_property
    def name(self):
        name = '(' + '|'.join([particle_dict[int(char)][0] for char in str(self.decayMT)[:-1] if char != '0']) + ')'
        name += f'm{self.decayMT%10}' if self.decayMT%10 != 0 else ''
        return name

    @cached_property
    def status(self):
        return self.decayMT % 10

    @cached_property
    def varid(self):
        varid = sum([particle_dict[int(char)][1] for char in str(self.decayMT)[:-1] if char != '0'])
        return (varid, self.status)

    def __repr__(self):
        return self.name

    def __getattr__(self, attr: str):
        if attr in self.data:
            return self.data[attr]
        else:
            return None


class Nuclide(list):

    def __init__(self, name: str, reaction: List[Union[Reaction, DecayReaction]] = [], data: dict = None):
        self.name = name
        self.data = data
        super().__init__(reaction)

    def __repr__(self):
        return self.name
    
    def __str__(self):
        return self.name + super().__repr__()

    def __getattr__(self, attr: str):
        if attr in self.data:
            return self.data[attr]
        else:
            return None

    @staticmethod
    def help():
        return "Nuclide(name, reaction=[Reaction, DecayReaction], data={'rate':float, 'Qalpha':float, 'Qbeta':float, 'Qgamma':float}"

    @cached_property
    def nucid(self):
        return decompose_nucname(self.name)["nucid"]

    @cached_property
    def mass(self):
        return decompose_nucname(self.name)["mass"]
    
    @cached_property
    def atomic(self):
        return decompose_nucname(self.name)["atomic"]

    @property
    def num_rec(self):
        return len(self)

    def __call__(self, reaction: Union[str, int]) -> Union[Reaction, DecayReaction]:
        rec_index = 'name' if isinstance(reaction, str) else 'MT'
        return next(reactioni for reactioni in self if getattr(reactioni, rec_index) == reaction)

    def sort_reactions(self):
        self.sort(key=lambda reaction: reaction.MT)

    def transmute_by_reaction(self, reaction: Union[Reaction, DecayReaction]):
        target = self.nucid + reaction.varid[0] * 10
        target = target // 10 * 10 + reaction.varid[1]
        return target
    
    def transmute_by_varid(self, varid: tuple):
        target = self.nucid + varid[0] * 10
        target = target // 10 * 10 + varid[1]
        return target

    def transmute(self):
        return [self.transmute_by_reaction(reaction) for reaction in self]

    def decay(self):
        return [self.transmute_by_reaction(decayreaction) for decayreaction in self]
    
class Nuclib(list):

    def __init__(self, type: str, nuclides: List[Nuclide] = []):
        self.type = type
        super().__init__(nuclides)

    def __call__(self, nuclide: Union[str, int]) -> Nuclide:
        """
        Returns the first instance of a nuclide in the class that matches the given nuclide name or ID.

        Parameters:
        - nuclide: A string representing the name of the nuclide or an integer representing the nuclide ID.

        Returns:
        - The first instance of the nuclide found in the class.

        Raises:
        - StopIteration: If no matching nuclide is found in the class.
        """
        nuc_index = 'name' if isinstance(nuclide, str) else 'nucid'
        return next(nuclidei for nuclidei in self if getattr(nuclidei, nuc_index) == nuclide)

    def sort_nuclides(self):
        self.sort(key=lambda nuclide: nuclide.nucid)
